--[[][][][][][][][][][][][][][][][][][][][][][]--
Functions to check if a given ship itemid matches a given loadout and vice versa
--[][][][][][][][][][][][][][][][][][][][][][]]--

local reconfig = {}

local STRICT_COLOR_TOL = 5 --max delta_E for strict color match
local LOOSE_COLOR_TOL = 15 --max delta_E for loose color match

--// Rank determined by sum of all applicable flags below. Lower numbers are better
local match_rankings = {
	--Ships
	SHIP_EXACT_MATCH = 0,
	SHIP_COMPATIBLE_VARIANT = 12, --variant & compatible
	SHIP_NON_VARIANT = 36, --non-variant & compatible

	--Addons
	ADDONS_EXACT_MATCH = 0,
	ADDONS_ADDITIONAL_PRESENT = 40,
	ADDONS_ALL_FROM_INVENTORY = 10,
	ADDONS_PURCHASE_REQUIRED = 24,
	ADDONS_MISSING_PER = 0.8, --this value multiplied by number of missing addons
	
	--Port Assignments
	PORTS_EXACT_MATCH = 0,
	PORTS_OUT_OF_ORDER = 12,

	--Weapon Groups
	WG_MATCH = 0,
	WG_DO_NOT_MATCH = 2,

	--Ship Colors
	COLOR_EXACT_MATCH = 0,
	COLOR_STRICT_MATCH = 1,
	COLOR_LOOSE_MATCH = 3,
	COLOR_DOES_NOT_MATCH = 8,
	
	--Ship Livery
	LIVERY_EXACT_MATCH = 0,
	LIVERY_DOES_NOT_MATCH = 9,
	
	--Other
	NO_POWERCELL = 6,
	CARGO_NOT_EMPTY = 15,
	AMMO_NEEDS_RELOAD = 18,
	SHIP_NEEDS_REPAIRS = 50,
}

--// Returns true if ships match based on ports & class per matching settings
--// a and be are ship loadouts
local ship_matching_check = {
	[1] = function(a,b) --exact match only
		return a==b
	end,
	[2] = function(a,b) --allow ship variants
		a = tonumber(a)
		b = tonumber(b)
		assert(a and b) --TODO error msg
		
		--TODO: need to check ports are compatible TOO
		
		--use ship icon to determine if ships are the same class
		return InvManager.GetInventoryItemParameter(a, 'Icon') == InvManager.GetInventoryItemParameter(b, 'Icon')
	end,
	[3] = function(a,b) --ports must be compatible
		--TODO need to get port types and count of ship itemid
		return true
	end,
	
}

--checks if itemtypeid from loadout is compatible with active ship (i.e. reconfigure active ship to match loadout)
local ship_match_active_check
ship_match_active_check = {
	[1] = function(loadout) --exact match only
		local loadout_itemtypeid = tonumber(loadout.ship)
		return HasActiveShip() and GetInventoryItemType(GetActiveShipID()) == loadout_itemtypeid
	end,
	[2] = function(loadout) --allow ship variants (ship icon must match AND ports must be compatible!!!)
		local itemtypeid = tonumber(loadout.ship)
		local is_compatible = HasActiveShip() and itemtypeid and GetInventoryItemIcon(GetActiveShipID()) == InvManager.GetInventoryItemParameter(itemtypeid, 'Icon')
		
		return is_compatible and ship_match_active_check[1](loadout)
	end,
	[3] = function(loadout) --ports must be compatible: active ship must have all ports present in loadout (extra ports allowed)
		if not HasActiveShip() then return end
		
		--count ports of each type on active ship
		local active_ports = {}
		for i=1,GetActiveShipNumAddonPorts() do
			local port_info = GetActiveShipPortInfo(i)
			local port_type = port_info.type
			active_ports[port_type] = (active_ports[port_type] or 0) + 1
		end
		
		--TODO: implement lines above at higher lever for efficiency
		
		local loadout_ports = {}
		for addon in ipairs(loadout.ports or {}) do
			local itemtypeid = tonumber(addon)
			if itemtypeid and itemtypeid>0 then
				local port_type = InvManager.GetInventoryItemPrototype(itemtypeid):GetClassSubType()
				loadout_ports[port_type] = (loadout_ports[port_type] or 0) + 1
			end
		end
		
		local is_compatible = true --assume true until proven false
		for port_type,count in pairs(loadout_ports) do
			local active_count = active_ports[port_type] or 0
			if loadout_ports > active_count then
				is_compatible = false
			end
		end
		
		return is_compatible
	end,
}	

--create table of each addon itemtypeid and its qty from loadout.ports
local function get_addons_from_ports(ports)
	local addon_list = {}
	for _,addon in pairs(ports or {}) do
		if addon>0 then
			if addon_list[addon] then
				addon_list[addon] = addon_list[addon] + 1
			else addon_list[addon] = 1 end
		end
	end
	
	return addon_list
end

--// Returns true if addons from b match a per matching settings
--// a and b are tables with itemids as keys and qty for values
local addon_matching_check = {
	[1] = function(a,b) --exact match only
		a = get_addons_from_ports(a)
		b = get_addons_from_ports(b)
		
		for itemid,qty in pairs(b) do
			if a[itemid]==qty then a[itemid]=nil else return false end --remove entry from b if qty matches a
		end
		return not next(a) --if b is now empty then is exact match
	end,
	[2] = function(a,b) --additional addons allowed
		a = get_addons_from_ports(a)
		b = get_addons_from_ports(b)
		
		for itemid,qty in pairs(b) do
			if not a[itemid] or a[itemid] < qty then return false end
		end
		return true			
	end,
	[3] = function(a,b) --load/buy missing addons and remove extras (i.e. reconfigure to match)
		--TODO missing addons must be available for purchase or in inventory
		return true
	end,
}

--// Returns true if port assignments from b match a per matching settings
--// a and b are indexed tables containing itemtypeids; last entry is zero (i.e. loadout.ports)
local port_matching_check = {
	[1] = function(a,b) --exact match only
		for i,itemtypeid in pairs(a) do
			if itemtypeid>0 and b[i] ~= itemtypeid then return false end
		end
		return true			
	end,
	[2] = function(a,b) --reconfigure to match
		return true --no conditions exist that prevent reconfiguring port assignments of b to match a
	end,
	[3] = function(a,b) --ignore port assignments
		return true
	end,
}

--// Returns true if weapon groups match per matching settings
--// a and b are an indexed table with 18 entries corresponding to each weapon trigger; each is a table with portids as keys and true if assigned to the given trigger
local wg_matching_check = {
	[1] = function(a,b) --exact match only
		for n=1,18 do --each trigger
			for port,_ in pairs(a) do
				if not b[port] then return false end
			end
			
			for port,_ in pairs(b) do
				if not a[port] then return false end
			end
		end
		return true
	end,
	[2] = function(a,b) --reconfigure to match
		return true --no conditions exist that prevent reconfiguring weapon groups of b to match a
	end,
	[3] = function(a,b) --ignore weapon groups
		return true
	end,	
}

--// Returns true if colors match per matching settings
--// a and b are color strings containing rgb values 0-255: e.g. "255 255 255"
local color_matching_check = {
	[1] = function(a,b) --exact match
		return a == b
	end,
	[2] = function(a,b) --strict match
		local dE = SLM.colors:calc_deltaE(a+1,b+1)
		return dE and dE <= STRICT_COLOR_TOL
	end,
	[3] = function(a,b) --loose match
		local dE = SLM.colors:calc_deltaE(a+1,b+1)
		return dE and dE <= LOOSE_COLOR_TOL
	end,
	[4] = function(a,b) --ignore color
		return true
	end,
}

--// Returns true if livery matches per matching settings
--// a and b are overlayindex
local livery_matching_check = {
	[1] = function(a,b) --exact match only
		return (a or 0)==(b or 0)
	end,
	[2] = function(a,b) --ignore livery
		return true
	end,
}

local cargo_matching_check = {
	[1] = function(itemid) --exclude ships with cargo
		local cargo = GetShipInventory(itemid).cargo or {}
		return not next(cargo)
	end,
	[2] = function(itemid) --unload cargo
		return true --TODO check if station is full full of cargo
	end,
}

local rr_matching_check = {
	[1] = function(itemid) --exclude ships needing repairs/reloads
		local current_armor, max_armor = GetAddonItemInfo(itemid)
		local is_needs_repairs = current_armor and current_armor<max_armor
		
		local ammo_prices = GetShipAmmoPrices(itemid) or {}
		local ammo_qty = ammo_prices.allammoquantity or 0
		
		return not is_needs_repairs and ammo_qty<=0
	end,
	[2] = function(itemid) --prompt to reload & repair
		return true --TODO check if ammo is reloadable? (e.g. seekers), probably better to just have warning
	end,
}

--which function and loadout key corresponds to each vars key for matching settings
local vars_names_lookup = {
	['MatchShipsBy'] = {ship_matching_check, 'ship'}, --loadout.ship
	['MatchAddonsBy'] = {addon_matching_check, 'ports'}, --loadout.ports
	['MatchPortsBy'] = {port_matching_check, 'ports'}, --loadout.ports
	['MatchWeaponGroupsBy'] = {wg_matching_check, 'groups'}, --loadout.groups
	['MatchColorsBy'] = {color_matching_check, 'shipcolor'}, --loadout.shipcolor
	['MatchLiveryBy'] = {livery_matching_check, 'overlayindex'}, --loadout.overlayindex
	['MatchCargoBy'] = {cargo_matching_check, nil}, --ship itemid
	['MatchReloadsRepairsBy'] = {rr_matching_check, nil}, --ship itemid
}

--// Calculate score for color matching
--// color1 & color2 are loadout.shipcolor
local function color_match_rank(color1, color2)
	color1 = tonumber(color1)
	color2 = tonumber(color2)
	assert(color1 and color2) --TODO error msg
	color1 = color1+1 --convert from shipcolor to pallette color
	color2 = color2+1 --convert from shipcolor to pallette color
	
	if color1==color2 then
		return match_rankings.COLOR_EXACT_MATCH
	else
		local dE = SLM.colors:calc_deltaE(color1, color2)
		if dE <= STRICT_COLOR_TOL then
			return match_rankings.COLOR_STRICT_MATCH
		elseif dE <= LOOSE_COLOR_TOL then
			return match_rankings.COLOR_LOOSE_MATCH
		else return match_rankings.COLOR_DOES_NOT_MATCH end
	end
end

--// Calculate score for port assignment matching
--// addons1 & addons2 are loadout.ports
local function port_match_rank(addons1, addons2)
	assert(type(addons1)=="table" and type(addons2)=="table") --TODO error msg
	return port_matching_check[2](addons1, addons2) and match_rankings.PORTS_EXACT_MATCH or match_rankings.PORTS_OUT_OF_ORDER
end

--// Calculate score for weapon group matching
--// wg1 & wg2 are loadout.groups
local function wg_match_rank(wg1, wg2)
	assert(type(wg1)=="table" and type(wg2)=="table") --TODO error msg
	return wg_matching_check[2](wg1, wg2) and match_rankings.WG_MATCH or match_rankings.WG_DO_NOT_MATCH
end

--// Calculate score for livery matching
--// overlay1 & overlay2 are loadout.overlayindex
local function livery_match_rank(overlay1, overlay2)
	return overlay1==overlay2 and match_rankings.LIVERY_EXACT_MATCH or match_rankings.LIVERY_DOES_NOT_MATCH
end

--// Extra rankings to calculate that don't depend on loadout
local function misc_match_rank(itemid)
	itemid = tonumber(itemid)
	assert(type(itemid)=="number") --TODO error msg
	
	local current_armor, max_armor = GetAddonItemInfo(itemid)
	local inv = GetShipInventory(itemid)
	
	local needs_reload = false --assume false until proven otherwise
	for _,addon in pairs(inv.addons or {}) do
		local current_ammo, max_ammo = GetAddonItemInfo(addon)
		if current_ammo and current_ammo < max_ammo then --both -1 if non-ammo addon
			needs_reload = true
			break
		end
	end 
	
	local score = GetShipCargoCount(itemid)>0 and match_rankings.CARGO_NOT_EMPTY or 0
	score = score + (current_armor and current_armor<max_armor and match_rankings.SHIP_NEEDS_REPAIRS or 0)
	score = score + (needs_reload and match_rankings.AMMO_NEEDS_RELOAD or 0)
	
	return score --higher numbers bad
end

--// Checks if specified ship itemid is compatible with specified loadout table per matching settings
function reconfig.is_ship_compatible(itemid, loadout)
	local ship_loadout = SaveShipLoadout(itemid) or {} --convert itemid to loadout
	
	local setting --index for vars matching settings
	for vars_key,info in pairs(vars_names_lookup) do --don't care about order
		setting = SLM.vars.GetVars(vars_key)
		
		local a
		if info[2] then a = ship_loadout[ info[2] ] else a = itemid end
		
		local b
		if info[2] then b = loadout[ info[2] ] end
		
		if not info[1][setting](a,b) then return false end --check each matching setting
	end
	
	return true
end

--// Calculates overall score from all matching categories
function reconfig.calc_rank(itemid, loadout)
	local ship_loadout = SaveShipLoadout(itemid) or {}
	if type(ship_loadout)~="table" then return end
	
	return color_match_rank(ship_loadout.shipcolor, loadout.shipcolor)
	 + port_match_rank(ship_loadout.ports or {}, loadout.ports or {})
	 + wg_match_rank(ship_loadout.groups, loadout.groups)
	 + livery_match_rank(ship_loadout.overlayindex or 0, loadout.overlayindex or 0)
	 + misc_match_rank(itemid)
end

--// Finds ship in inventory that is best match (lowest score) to specified loadout table
function reconfig.best_local_ship(loadout)
	if type(loadout)~="table" then return end
	if not PlayerInStation() then return end
	
	local active_itemid = HasActiveShip() and GetActiveShipID()
	
	local scores = {} --score for each compatible itemid in local inventory
	for _, itemid in ipairs(GetShipList() or {}) do --for each ship in local inventory
		if reconfig.is_ship_compatible(itemid, loadout) then
			local score = reconfig.calc_rank(itemid, loadout)
			
			--console_print("SCORE: "..tostring(score))
			
			if score == 0 then --score of 0 can't be beat
				if itemid~=active_itemid then --can't switch to active ship
					return itemid, score
				else return end
			end 
			
			scores[itemid] = score
		end
	end
	
	local best_score
	local best_ship
	for itemid,score in pairs(scores) do
		if best_score and score<best_score then
			best_score = score
			best_ship = itemid
		else  --first entry
			best_score = score
			best_ship = itemid
		end
	end
	
	if best_ship==active_itemid then return end --can't switch to active ship
	
	return best_ship, best_score
end

--// Reconfigures the active ship to specified loadout table per current matching settings
--// Presently used just after switching to ship from inventory matching loadout
function reconfig.ReconfigureActiveShip(loadout)
	if type(loadout)~='table' then return end --TODO assert
	if not HasActiveShip() then return end --TODO assert
	
	NotificationDialog:SetMessage("Reconfiguring...")
	ShowDialog(NotificationDialog, iup.CENTER, iup.CENTER)
	
	local active_ship = GetActiveShipID()
	local ship_loadout = SaveShipLoadout(active_ship) or {} --convert itemid to loadout
	
	--Assume correct addons present --TODO
	--Not currently possible to change ship color
	--Not currently possible to change ship livery
	
	--Reconfigure port assignments to match loadout	
	assert(type(loadout.ports)=='table') --TODO error msg
	local to_disconnect = {}
	local to_connect = {}
	local avail_addons = {}
	
	--make list of addons to disconnect
	for n=1,GetActiveShipNumAddonPorts() do
		local addon_itemid = GetActiveShipItemIDAtPort(n)
		local addon_itemtypeid = GetInventoryItemType(addon_itemid)
		local loadout_itemtypeid = loadout.ports[n]
		
		if loadout_itemtypeid and loadout_itemtypeid>0 then
			if addon_itemtypeid ~= loadout_itemtypeid then
				if addon_itemtypeid then
					table.insert(to_disconnect, addon_itemid)
				
					if not avail_addons[addon_itemtypeid] then avail_addons[addon_itemtypeid]={} end
					table.insert(avail_addons[addon_itemtypeid], addon_itemid)
				end
				
				table.insert(to_connect, {loadout_itemtypeid, n}) --use itemtypeid for now, change to itemid next pass
			end
		elseif addon_itemtypeid then --ship addon this port needs to be removed if present
			table.insert(to_disconnect, addon_itemid)
				
			if not avail_addons[addon_itemtypeid] then avail_addons[addon_itemtypeid]={} end
			table.insert(avail_addons[addon_itemtypeid], addon_itemid)
		end
	end	
	
	--make list of cargo to unload
	local cargo = GetShipInventory(active_ship).cargo or {}
	local to_unload = {}
	for _,cargo_itemid in ipairs(cargo) do
		local qty = GetInventoryItemQuantity(cargo_itemid)
		if qty then table.insert(to_unload, {itemid=cargo_itemid, quantity=qty}) end
	end
	
	--reload and repair
	local function reload_repair_cb()
		local ammo_prices = GetShipAmmoPrices(active_ship) or {}
		local ammo_qty = ammo_prices.allammoquantity or 0
		local ammo_cost = ammo_prices.allammoprice or 0
		
		local ship_health = {GetActiveShipHealth()}
		local is_needs_repairs = ship_health[7] and ship_health[7]>0
		local repair_info = GetStationAmmoInfoByID(active_ship) or {}
		local repair_cost = repair_info.price or 0
		
		local total_cost = ammo_cost + repair_cost
		
		--prompt player for reloads/repairs if needed
		if ammo_qty>0 or is_needs_repairs then
			local function repair_cb()
				if is_needs_repairs then
					RepairShip(active_ship, 1, function(err)
						if not err then
							purchaseprint("Ship repaired for a total price of "..(comma_value(repair_cost) or "???").."c.")
						end
					end)
				end
				HideDialog(QuestionDialog)
				HideDialog(NotificationDialog)
			end
			
			QuestionDialog:SetMessage(
				string.format("Warning: The active ship needs to be %s%s%s. The cost is %dc. How do you want to proceed?",
					ammo_qty>0 and "reloaded" or "",
					ammo_qty>0 and is_needs_repairs and " and " or "",
					is_needs_repairs and "repaired" or "",
					total_cost),
				string.format("%s%s%s", --"Reload & Repair"
					ammo_qty>0 and "Reload" or "",
					ammo_qty>0 and is_needs_repairs and " & " or "",
					is_needs_repairs and "Repair" or ""),
				function()
					if ammo_qty>0 then
						ReplenishAll(active_ship, function(err)
							if not err then
								purchaseprint("All ammo in active ship purchased for a total price of "..(comma_value(ammo_cost) or "???").."c.")
							end
							repair_cb()
						end)
					else repair_cb() end
				end,
				"Cancel", function() HideDialog(QuestionDialog) HideDialog(NotificationDialog) end) --do nothing
			ShowDialog(QuestionDialog, iup.CENTER, iup.CENTER)
		else HideDialog(NotificationDialog) end
	end
	
	--re-arrange addons to correct ports
	local swap_addons_cb
	swap_addons_cb = function()
		local itemtypeid
		local itemid, port = table.remove(to_disconnect)
		
		if itemid then --disconnect addons first
			DisconnectAddon(itemid, swap_addons_cb)
		else --then connect addons
			itemtypeid, port = unpack(table.remove(to_connect) or {})
			if itemtypeid then
				for _,itemid in ipairs(GetStationAddonList() or {}) do
					if GetInventoryItemType(itemid)==itemtypeid then
						ConnectAddon(port, itemid, swap_addons_cb)
						break
					end
				end	
			else reload_repair_cb() end --addon swap done
		end
	end
	
	--unload any cargo
	local function unload_cargo_cb()
		if next(to_unload) then
			CheckStorageAndUnloadCargo(to_unload, swap_addons_cb)
		else swap_addons_cb() end
	end
	
	--Reconfigure weapons groups to match loadout
	assert(type(loadout.groups)=='table') --TODO error msg
	ConfigureMultipleWeaponGroups(loadout.groups, unload_cargo_cb)
end

return reconfig